/*
 * Copyright (C) 2001 USAGI/WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <netinet/in.h>
#include <netdb.h>
#include <getopt.h>

#ifdef TCP_MD5SIG
# define ENABLE_MD5SIG
#endif

#ifdef ENABLE_MD5SIG
# if HAVE_LINUX_TCP_H
#  include <linux/tcp.h>
# endif
#include "base64.h"
#endif

int main(int argc, char **argv){
	const char *myname = argv[0];
	struct msghdr msghdr;
	struct iovec iov[2];
	char buf[128];
	struct sockaddr_storage sa;
	int salen;
	int s, new;
	int c;
	int v6only, setv6only = 0;
	struct addrinfo hints, *res0;
	int gai;
	char *host, *serv;
	char abuf[NI_MAXHOST];
	int loop = 0;
#ifdef ENABLE_MD5SIG
	char *key = NULL;
	int keylen = 0;
#endif
	int reuse = 0;

	while((c = getopt(argc, argv, "6:a"
#ifdef ENABLE_MD5SIG
				      "k:"
#endif
				      "l"
			  )) != EOF) {
		switch(c) {
		case '6':
			setv6only = 1;
			v6only = atoi(optarg);
			break;
		case 'a':
			reuse = 1;
			break;
#ifdef ENABLE_MD5SIG
		case 'k':
		    {
			int len = strlen(optarg);
			if (len < 0)
				goto help;
			free(key);
			key = malloc(b64_space(len));
			if (!key) {
				perror("malloc");
				exit(1);
			}
			keylen = b64_decode(key, optarg, len);
			if (keylen < 0) {
				printf("%s: base64 decode error: %s\n", myname, optarg);
				exit(1);
			}
			break;
		    }
#endif
		case 'l':
			loop = 1;
			break;
		}
	}

	argc -= optind;
	argv += optind;

	if (argc == 0) {
help:
		printf("%s [-6 v6only-sockopt] [-k base64-md5-key] [host] port\n", myname);
		exit(1);
	}
        host = argc > 1 ? argv[0] : NULL;
        serv = argc > 1 ? argv[1] : argv[0];

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = PF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags = AI_PASSIVE;
	gai = getaddrinfo(host, serv, &hints, &res0);
	if (gai) {
		printf("getaddrinfo(): %s\n",
			gai_strerror(gai));
		exit(2);
	}

	s = socket(res0->ai_family, res0->ai_socktype, res0->ai_protocol);
	if (s < 0) {
		perror("socket");
		exit(2);
	}

	if (setv6only) {
		printf("set IPV6_V6ONLY to %d\n", v6only);
#ifdef IPV6_V6ONLY
		if (setsockopt(s, IPPROTO_IPV6, IPV6_V6ONLY, &v6only, sizeof(v6only)) < 0)
			perror("setsockopt(IPV6_V6ONLY)");
#else
		fprintf(stderr, "IPV6_V6ONLY not supported.\n");
#endif
	}

	if (reuse) {
		if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &reuse, sizeof(reuse)) < 0)
			perror("setsockopt(SO_REUSEADDR)");
	}

#ifdef ENABLE_MD5SIG
	if (key) {
# ifdef TCP_MD5SIG
#  ifndef TCP_MD5SIG_KEY
		/* FreeBSD / NetBSD */
		int md5sig = 1;
#  else
		/* Linux */
		struct tcp_md5sig_key md5sig;
#  endif
# elif defined(TCP_RFC2385)
		struct tcp_rfc2385_cmd md5sig;
# endif
# if defined(TCP_MD5SIG_KEY) || defined(TCP_RFC2385)
		/*XXX*/
		char nhost[NI_MAXHOST];
		struct addrinfo *peer;
		getnameinfo(res0->ai_addr, res0->ai_addrlen, nhost, sizeof(nhost), NULL, 0, NI_NUMERICHOST);
		hints.ai_flags = AI_NUMERICHOST;
		getaddrinfo(nhost, NULL, &hints, &peer);

		memset(&md5sig, 0, sizeof(md5sig));
#  ifdef TCP_MD5SIG_KEY
		md5sig.tcpsig_op = TCP_MD5SIG_KEY;
		md5sig.tcpsig_key = key;
		md5sig.tcpsig_keylen = keylen;
		md5sig.tcpsig_peerlen = peer->ai_addrlen;
		md5sig.tcpsig_peer = peer->ai_addr;
#  else
		if (peer->ai_addr->sa_family != AF_INET) {
			perror("unsupported family for MD5 Signature");
			exit(1);
		}
		md5sig.command = TCP_RFC2385_ADD;
		md5sig.address = ((struct sockaddr_in *)peer->ai_addr)->sin_addr.s_addr;	/*XXX*/
		md5sig.keylen = keylen;
		md5sig.key = key;
#  endif	/* TCP_MD5SIG_KEY */
# endif	/* defined(TCP_MD5SIG_KEY) || defined(TCP_RFC2385) */
		if (setsockopt(s, IPPROTO_TCP, TCP_MD5SIG, &md5sig, sizeof(md5sig)) < 0)
			perror("setsockopt(TCP_MD5SIG)");
# if defined(TCP_MD5SIG_KEY) || defined(TCP_RFC2385)
		freeaddrinfo(peer);
# endif
	}
#endif

	if (bind(s, res0->ai_addr, res0->ai_addrlen)) {
		freeaddrinfo(res0);
		perror("bind");
		exit(2);
	}
	freeaddrinfo(res0);

	listen(s, 1);

	salen = sizeof(sa);
	new = accept(s, (struct sockaddr *)&sa, &salen);
	
	if (new < 0) {
		perror("accept");
		exit(2);
	}

	if (getnameinfo((struct sockaddr *)&sa, salen,
			abuf, sizeof(abuf), NULL, 0,
			NI_NUMERICHOST))
		strcpy(abuf, "???");

	printf("connected from %s\n", abuf);

	if (reuse) {
		int reuse_size = sizeof(reuse);

		if (getsockopt(new, SOL_SOCKET, SO_REUSEADDR, &reuse, &reuse_size) < 0)
			perror("getsockopt(SO_REUSEADDR)");
		else
			printf("newfd(%d; old=%d)'s REUSEADDR => %d\n", new, s, reuse);
	}

	while(loop) {
		int c = read(new, buf, sizeof(buf));
		printf("read = %d:", c);
		if (c > 0) {
			int i;
			write(new, buf, c);
			for (i = 0; i < c; i++)
				printf(" %02x", buf[i] & 0xff);
			printf("\n");
		}
	}

	exit(0);
}

